"""
A cython implementation of the bounded priority queue

This is a priority queue that only keeps track of smallest k values that have
been added to it.

This priority queue is implemented with the configuration of having the largest
element at the beginning - this exploited to store nearest neighbour lists.

"""


import numpy as np

cimport cython
cimport numpy as np
from cpython.mem cimport PyMem_Free, PyMem_Malloc, PyMem_Realloc


cdef class BoundedPriorityQueue:
    def __cinit__(self, np.intp_t max_elements, np.intp_t pids=0):
        self.max_elements = max_elements
        # mark invalid recently  values with -1
        self.heap = np.zeros(max_elements)-1
        self.heap_ptr = &(self.heap[0])
        # only allocate memory if we intend to store particle ID's
        self.use_pids = pids
        if pids == 1:
            self.pids = np.zeros(max_elements, dtype="int64")-1
            self.pids_ptr = &(self.pids[0])

        self.size = 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int max_heapify(self, np.intp_t index) nogil except -1:
        cdef np.intp_t left = 2 * index + 1
        cdef np.intp_t right = 2 * index + 2
        cdef np.intp_t largest = index

        if left < self.size and self.heap_ptr[left] > self.heap_ptr[largest]:
            largest = left
        if right < self.size and self.heap_ptr[right] > self.heap_ptr[largest]:
            largest = right

        if largest != index:
            self.heap_ptr[index], self.heap_ptr[largest] = \
                self.heap_ptr[largest], self.heap_ptr[index]
            if self.use_pids:
                self.pids_ptr[index], self.pids_ptr[largest] = \
                    self.pids_ptr[largest], self.pids_ptr[index]

            self.max_heapify(largest)

        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int propagate_up(self, np.intp_t index) nogil except -1:
        while index != 0 and self.heap_ptr[(index - 1) // 2] < self.heap_ptr[index]:
            self.heap_ptr[index], self.heap_ptr[(index - 1) // 2] = self.heap_ptr[(index - 1) // 2], self.heap_ptr[index]
            if self.use_pids:
                self.pids_ptr[index], self.pids_ptr[(index - 1) // 2] = self.pids_ptr[(index - 1) // 2], self.pids_ptr[index]
            index = (index - 1) // 2

        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int add(self, np.float64_t val) nogil except -1:
        # if not at max size append, if at max size, only append if smaller than
        # the maximum value
        if self.size == self.max_elements:
            if val < self.heap_ptr[0]:
                self.extract_max()
                self.heap_append(val, -1)
        else:
            self.heap_append(val, -1)
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int add_pid(self, np.float64_t val, np.int64_t ind) nogil except -1:
        if self.size == self.max_elements:
            if val < self.heap_ptr[0]:
                self.extract_max()
                self.heap_append(val, ind)
        else:
            self.heap_append(val, ind)
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int heap_append(self, np.float64_t val, np.int64_t ind) nogil except -1:
        self.heap_ptr[self.size] = val
        if self.use_pids:
            self.pids_ptr[self.size] = ind
        self.size += 1
        self.propagate_up(self.size - 1)
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef np.float64_t extract_max(self) nogil except -1:
        cdef np.float64_t maximum = self.heap_ptr[0]
        cdef np.float64_t val
        cdef np.int64_t ind

        val = self.heap_ptr[self.size-1]
        self.heap_ptr[self.size-1] = -1

        if self.use_pids:
            ind = self.pids_ptr[self.size-1]
            self.pids_ptr[self.size-1] = -1

        self.size -= 1
        if self.size > 0:
            self.heap_ptr[0] = val
            if self.use_pids:
                self.pids_ptr[0] = ind
            self.max_heapify(0)
        return maximum

    cdef int validate_heap(self) nogil except -1:
        # this function loops through every element in the heap, if any children
        # are greater than their parents then we return zero, which is an error
        # as the heap condition is not satisfied
        cdef int i, index
        for i in range(self.size-1, -1, -1):
            index = i
            while index != 0:
                if self.heap_ptr[index] > self.heap_ptr[(index - 1) // 2]:
                    return 0
                index = (index - 1) // 2
        return 1

cdef class NeighborList:
    def __cinit__(self, np.intp_t init_size=32):
        self.size = 0
        self._max_size = init_size
        self.data_ptr = <np.float64_t*> PyMem_Malloc(
            self._max_size * sizeof(np.float64_t)
        )
        self.pids_ptr = <np.int64_t*> PyMem_Malloc(
            self._max_size * sizeof(np.int64_t)
        )
        self._update_memview()

    def __dealloc__(self):
        PyMem_Free(self.data_ptr)
        PyMem_Free(self.pids_ptr)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int _update_memview(self) except -1:
        self.data = <np.float64_t[:self._max_size]> self.data_ptr
        self.pids = <np.int64_t[:self._max_size]> self.pids_ptr

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int _extend(self) nogil except -1:
        if self.size == self._max_size:
            self._max_size *= 2
            with gil:
                self.data_ptr = <np.float64_t*> PyMem_Realloc(
                    self.data_ptr,
                    self._max_size * sizeof(np.float64_t)
                )
                self.pids_ptr = <np.int64_t*> PyMem_Realloc(
                    self.pids_ptr,
                    self._max_size * sizeof(np.int64_t)
                )
                self._update_memview()
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    @cython.initializedcheck(False)
    cdef int add_pid(self, np.float64_t val, np.int64_t ind) nogil except -1:
        self._extend()
        self.data_ptr[self.size] = val
        self.pids_ptr[self.size] = ind
        self.size += 1
        return 0

# these are test functions which are called from
# yt/utilities/lib/tests/test_nn.py
# they are stored here to allow easy interaction with functions not exposed at
# the python level
def validate_pid():
    m = BoundedPriorityQueue(5, True)

    # Add elements to the queue
    elements = [0.1, 0.25, 1.33, 0.5, 3.2, 4.6, 2.0, 0.4, 4.0, .001]
    pids = [1,2,3,4,5,6,7,8,9,10]

    for el, pid in zip(elements, pids):
        m.add_pid(el, pid)

    m.extract_max()
    m.extract_max()
    m.extract_max()

    return np.asarray(m.heap), np.asarray(m.pids)

def validate():
    m = BoundedPriorityQueue(5)

    # Add elements to the queue
    for el in [0.1, 0.25, 1.33, 0.5, 3.2, 4.6, 2.0, 0.4, 4.0, .001]:
        m.add(el)

    m.extract_max()
    m.extract_max()
    m.extract_max()

    return np.asarray(m.heap)

def validate_nblist():
    nblist = NeighborList(init_size=2)

    for i in range(4):
        nblist.add_pid(1.0, i)

    # Copy is necessary here. Without it, the allocated memory would be freed.
    # Leaving random data array.
    return np.asarray(nblist.data).copy(), np.asarray(nblist.pids).copy()



import subprocess
import sys

from docutils.parsers.rst import Directive


def setup(app):
    app.add_directive("yt_showfields", ShowFields)
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    retdict = dict(version="1.0", parallel_read_safe=True, parallel_write_safe=True)

    return retdict


class ShowFields(Directive):
    required_arguments = 0
    optional_arguments = 0
    parallel_read_safe = True
    parallel_write_safe = True

    def run(self):
        rst_file = self.state_machine.document.attributes["source"]
        lines = subprocess.check_output(
            [sys.executable, "./helper_scripts/show_fields.py"]
        )
        lines = lines.decode("utf8")
        lines = lines.split("\n")
        self.state_machine.insert_input(lines, rst_file)
        return []



# A namespace to store simple backports from most recent Python versions
# Backports should be contained in (if/else) blocks, cheking on the runtime
# Python version, e.g.
#
# if sys.version_info < (3, 8):
#     ... # insert backported definitions
# else:
#     pass
#
# while the else part if meant as a no-op, it is required to allow automated
# cleanups with pyupgrade, when our minimal supported Python version reaches
# the requirement for backports to be unneeded.
#
# Likewise, imports from this module should be guarded by a similar condition,
# e.g.,
#
# if sys.version_info >= (3, 8):
#     from functools import cached_property
# else:
#     from yt._maintenance.backports import cached_property
import sys

if sys.version_info < (3, 8):
    from _thread import RLock  # type: ignore
    from typing import List

    GenericAlias = type(List[int])

    _NOT_FOUND = object()

    class cached_property:
        def __init__(self, func):
            self.func = func
            self.attrname = None
            self.__doc__ = func.__doc__
            self.lock = RLock()

        def __set_name__(self, owner, name):
            if self.attrname is None:
                self.attrname = name
            elif name != self.attrname:
                raise TypeError(
                    "Cannot assign the same cached_property to two different names "
                    f"({self.attrname!r} and {name!r})."
                )

        def __get__(self, instance, owner=None):
            if instance is None:
                return self
            if self.attrname is None:
                raise TypeError(
                    "Cannot use cached_property instance without calling __set_name__ on it."
                )
            try:
                cache = instance.__dict__
            except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
                msg = (
                    f"No '__dict__' attribute on {type(instance).__name__!r} "
                    f"instance to cache {self.attrname!r} property."
                )
                raise TypeError(msg) from None
            val = cache.get(self.attrname, _NOT_FOUND)
            if val is _NOT_FOUND:
                with self.lock:
                    # check if another thread filled cache while we awaited lock
                    val = cache.get(self.attrname, _NOT_FOUND)
                    if val is _NOT_FOUND:
                        val = self.func(instance)
                        try:
                            cache[self.attrname] = val
                        except TypeError:
                            msg = (
                                f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                                f"does not support item assignment for caching {self.attrname!r} property."
                            )
                            raise TypeError(msg) from None
            return val

        __class_getitem__ = classmethod(GenericAlias)

else:
    pass


builtin_zip = zip
if sys.version_info >= (3, 10):
    zip = builtin_zip
else:
    # this function is deprecated in more_itertools
    # because it is superseded by the standard library
    from more_itertools import zip_equal

    def zip(*args, strict=False):
        if strict:
            return zip_equal(*args)
        else:
            return builtin_zip(*args)



# Using AMRKDTree Homogenized Volumes to examine large datasets
# at lower resolution.

# In this example we will show how to use the AMRKDTree to take a simulation
# with 8 levels of refinement and only use levels 0-3 to render the dataset.

# Currently this cookbook is flawed in that the data that is covered by the
# higher resolution data gets masked during the rendering.  This should be
# fixed by changing either the data source or the code in
# yt/utilities/amr_kdtree.py where data is being masked for the partitioned
# grid.  Right now the quick fix is to create a data_collection, but this
# will only work for patch based simulations that have ds.index.grids.

# We begin by loading up yt, and importing the AMRKDTree
import numpy as np

import yt
from yt.utilities.amr_kdtree.api import AMRKDTree

# Load up a dataset and define the kdtree
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
im, sc = yt.volume_render(ds, ("gas", "density"), fname="v0.png")
sc.camera.set_width(ds.arr(100, "kpc"))
render_source = sc.get_source()
kd = render_source.volume

# Print out specifics of KD Tree
print("Total volume of all bricks = %i" % kd.count_volume())
print("Total number of cells = %i" % kd.count_cells())

new_source = ds.all_data()
new_source.max_level = 3
kd_low_res = AMRKDTree(ds, data_source=new_source)
print(kd_low_res.count_volume())
print(kd_low_res.count_cells())

# Now we pass this in as the volume to our camera, and render the snapshot
# again.

render_source.set_volume(kd_low_res)
render_source.set_field(("gas", "density"))
sc.save("v1.png", sigma_clip=6.0)

# This operation was substantially faster.  Now lets modify the low resolution
# rendering until we find something we like.

tf = render_source.transfer_function
tf.clear()
tf.add_layers(
    4,
    0.01,
    col_bounds=[-27.5, -25.5],
    alpha=np.ones(4, dtype="float64"),
    colormap="RdBu_r",
)
sc.save("v2.png", sigma_clip=6.0)

# This looks better.  Now let's try turning on opacity.

tf.grey_opacity = True
sc.save("v3.png", sigma_clip=6.0)
#
## That seemed to pick out some interesting structures.  Now let's bump up the
## opacity.
#
tf.clear()
tf.add_layers(
    4,
    0.01,
    col_bounds=[-27.5, -25.5],
    alpha=10.0 * np.ones(4, dtype="float64"),
    colormap="RdBu_r",
)
tf.add_layers(
    4,
    0.01,
    col_bounds=[-27.5, -25.5],
    alpha=10.0 * np.ones(4, dtype="float64"),
    colormap="RdBu_r",
)
sc.save("v4.png", sigma_clip=6.0)
#
## This looks pretty good, now lets go back to the full resolution AMRKDTree
#
render_source.set_volume(kd)
sc.save("v5.png", sigma_clip=6.0)

# This looks great!



import os
import weakref

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.funcs import mylog, setdefaultattr
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.cosmology import Cosmology
from yt.utilities.file_handler import HDF5FileHandler

from .definitions import geometry_parameters
from .fields import GAMERFieldInfo


class GAMERGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id, filename=index.index_filename, index=index)
        self.Parent = None  # do NOT initialize Parent as []
        self.Children = []
        self.Level = level


class GAMERHierarchy(GridIndex):
    grid = GAMERGrid
    _preload_implemented = True  # since gamer defines "_read_chunk_data" in io.py

    def __init__(self, ds, dataset_type="gamer"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        self._handle = ds._handle
        self._group_grid = ds._group_grid
        self._group_particle = ds._group_particle
        self.float_type = "float64"  # fixed even when FLOAT8 is off
        self._particle_handle = ds._particle_handle
        self.refine_by = ds.refine_by
        self.pgroup = self.refine_by**3  # number of patches in a patch group
        GridIndex.__init__(self, ds, dataset_type)

    def _detect_output_fields(self):
        # find all field names in the current dataset
        # grid fields
        self.field_list = [("gamer", v) for v in self._group_grid.keys()]

        # particle fields
        if self._group_particle is not None:
            self.field_list += [("io", v) for v in self._group_particle.keys()]

    def _count_grids(self):
        # count the total number of patches at all levels
        self.num_grids = self.dataset.parameters["NPatch"].sum() // self.pgroup

    def _parse_index(self):
        parameters = self.dataset.parameters
        gid0 = 0
        grid_corner = self._handle["Tree/Corner"][()][:: self.pgroup]
        convert2physical = self._handle["Tree/Corner"].attrs["Cvt2Phy"]

        self.grid_dimensions[:] = parameters["PatchSize"] * self.refine_by

        for lv in range(0, parameters["NLevel"]):
            num_grids_level = parameters["NPatch"][lv] // self.pgroup
            if num_grids_level == 0:
                break

            patch_scale = (
                parameters["PatchSize"] * parameters["CellScale"][lv] * self.refine_by
            )

            # set the level and edge of each grid
            # (left/right_edge are YT arrays in code units)
            self.grid_levels.flat[gid0 : gid0 + num_grids_level] = lv
            self.grid_left_edge[gid0 : gid0 + num_grids_level] = (
                grid_corner[gid0 : gid0 + num_grids_level] * convert2physical
            )
            self.grid_right_edge[gid0 : gid0 + num_grids_level] = (
                grid_corner[gid0 : gid0 + num_grids_level] + patch_scale
            ) * convert2physical

            gid0 += num_grids_level
        self.grid_left_edge += self.dataset.domain_left_edge
        self.grid_right_edge += self.dataset.domain_left_edge

        # allocate all grid objects
        self.grids = np.empty(self.num_grids, dtype="object")
        for i in range(self.num_grids):
            self.grids[i] = self.grid(i, self, self.grid_levels.flat[i])

        # maximum level with patches (which can be lower than MAX_LEVEL)
        self.max_level = self.grid_levels.max()

        # number of particles in each grid
        try:
            self.grid_particle_count[:] = np.sum(
                self._handle["Tree/NPar"][()].reshape(-1, self.pgroup), axis=1
            )[:, None]
        except KeyError:
            self.grid_particle_count[:] = 0.0

        # calculate the starting particle indices for each grid (starting from 0)
        # --> note that the last element must store the total number of particles
        #    (see _read_particle_coords and _read_particle_fields in io.py)
        self._particle_indices = np.zeros(self.num_grids + 1, dtype="int64")
        np.add.accumulate(
            self.grid_particle_count.squeeze(), out=self._particle_indices[1:]
        )

    def _populate_grid_objects(self):
        son_list = self._handle["Tree/Son"][()]

        for gid in range(self.num_grids):
            grid = self.grids[gid]
            son_gid0 = (
                son_list[gid * self.pgroup : (gid + 1) * self.pgroup] // self.pgroup
            )

            # set up the parent-children relationship
            grid.Children = [self.grids[t] for t in son_gid0[son_gid0 >= 0]]

            for son_grid in grid.Children:
                son_grid.Parent = grid

            # set up other grid attributes
            grid._prepare_grid()
            grid._setup_dx()

        # validate the parent-children relationship in the debug mode
        if self.dataset._debug:
            self._validate_parent_children_relationship()

    # for _debug mode only
    def _validate_parent_children_relationship(self):
        mylog.info("Validating the parent-children relationship ...")

        father_list = self._handle["Tree/Father"][()]

        for grid in self.grids:
            # parent->children == itself
            if grid.Parent is not None:
                assert (
                    grid in grid.Parent.Children
                ), "Grid %d, Parent %d, Parent->Children[0] %d" % (
                    grid.id,
                    grid.Parent.id,
                    grid.Parent.Children[0].id,
                )

            # children->parent == itself
            for c in grid.Children:
                assert c.Parent is grid, "Grid %d, Children %d, Children->Parent %d" % (
                    grid.id,
                    c.id,
                    c.Parent.id,
                )

            # all refinement grids should have parent
            if grid.Level > 0:
                assert (
                    grid.Parent is not None and grid.Parent.id >= 0
                ), "Grid %d, Level %d, Parent %d" % (
                    grid.id,
                    grid.Level,
                    grid.Parent.id if grid.Parent is not None else -999,
                )

            # parent index is consistent with the loaded dataset
            if grid.Level > 0:
                father_gid = father_list[grid.id * self.pgroup] // self.pgroup
                assert (
                    father_gid == grid.Parent.id
                ), "Grid %d, Level %d, Parent_Found %d, Parent_Expect %d" % (
                    grid.id,
                    grid.Level,
                    grid.Parent.id,
                    father_gid,
                )

            # edges between children and parent
            for c in grid.Children:
                for d in range(0, 3):
                    msgL = (
                        "Grid %d, Child %d, Grid->EdgeL %14.7e, Children->EdgeL %14.7e"
                        % (grid.id, c.id, grid.LeftEdge[d], c.LeftEdge[d])
                    )
                    msgR = (
                        "Grid %d, Child %d, Grid->EdgeR %14.7e, Children->EdgeR %14.7e"
                        % (grid.id, c.id, grid.RightEdge[d], c.RightEdge[d])
                    )
                    if not grid.LeftEdge[d] <= c.LeftEdge[d]:
                        raise ValueError(msgL)

                    if not grid.RightEdge[d] >= c.RightEdge[d]:
                        raise ValueError(msgR)

        mylog.info("Check passed")


class GAMERDataset(Dataset):
    _index_class = GAMERHierarchy
    _field_info_class = GAMERFieldInfo
    _handle = None
    _group_grid = None
    _group_particle = None
    _debug = False  # debug mode for the GAMER frontend

    def __init__(
        self,
        filename,
        dataset_type="gamer",
        storage_filename=None,
        particle_filename=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):

        if self._handle is not None:
            return

        self.fluid_types += ("gamer",)
        self._handle = HDF5FileHandler(filename)
        self.particle_filename = particle_filename

        # to catch both the new and old data formats for the grid data
        try:
            self._group_grid = self._handle["GridData"]
        except KeyError:
            self._group_grid = self._handle["Data"]

        if "Particle" in self._handle:
            self._group_particle = self._handle["Particle"]

        if self.particle_filename is None:
            self._particle_handle = self._handle
        else:
            self._particle_handle = HDF5FileHandler(self.particle_filename)

        # currently GAMER only supports refinement by a factor of 2
        self.refine_by = 2

        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        self.storage_filename = storage_filename

    def _set_code_unit_attributes(self):
        if self.parameters["Opt__Unit"]:
            # GAMER units are always in CGS
            setdefaultattr(
                self, "length_unit", self.quan(self.parameters["Unit_L"], "cm")
            )
            setdefaultattr(self, "mass_unit", self.quan(self.parameters["Unit_M"], "g"))
            setdefaultattr(self, "time_unit", self.quan(self.parameters["Unit_T"], "s"))

            if self.mhd:
                setdefaultattr(
                    self, "magnetic_unit", self.quan(self.parameters["Unit_B"], "gauss")
                )

        else:
            if len(self.units_override) == 0:
                mylog.warning(
                    "Cannot determine code units ==> "
                    "Use units_override to specify the units"
                )

            for unit, value, cgs in [
                ("length", 1.0, "cm"),
                ("time", 1.0, "s"),
                ("mass", 1.0, "g"),
                ("magnetic", np.sqrt(4.0 * np.pi), "gauss"),
            ]:
                setdefaultattr(self, f"{unit}_unit", self.quan(value, cgs))

                if len(self.units_override) == 0:
                    mylog.warning("Assuming %8s unit = %f %s", unit, value, cgs)

    def _parse_parameter_file(self):

        # code-specific parameters
        for t in self._handle["Info"]:
            info_category = self._handle["Info"][t]
            for v in info_category.dtype.names:
                self.parameters[v] = info_category[v]

        # shortcut for self.parameters
        parameters = self.parameters

        # reset 'Model' to be more readable
        # (no longer regard MHD as a separate model)
        if parameters["Model"] == 1:
            parameters["Model"] = "Hydro"
        elif parameters["Model"] == 3:
            parameters["Model"] = "ELBDM"
        else:
            parameters["Model"] = "Unknown"

        # simulation time and domain
        self.dimensionality = 3  # always 3D
        self.domain_left_edge = parameters.get(
            "BoxEdgeL", np.array([0.0, 0.0, 0.0])
        ).astype("f8")
        self.domain_right_edge = parameters.get(
            "BoxEdgeR", parameters["BoxSize"]
        ).astype("f8")
        self.domain_dimensions = parameters["NX0"].astype("int64")

        # periodicity
        if parameters["FormatVersion"] >= 2106:
            periodic_bc = 1
        else:
            periodic_bc = 0
        self._periodicity = (
            bool(parameters["Opt__BC_Flu"][0] == periodic_bc),
            bool(parameters["Opt__BC_Flu"][2] == periodic_bc),
            bool(parameters["Opt__BC_Flu"][4] == periodic_bc),
        )

        # cosmological parameters
        if parameters["Comoving"]:
            self.cosmological_simulation = 1
            # here parameters["Time"][0] is the scale factor a at certain redshift
            self.current_redshift = 1.0 / parameters["Time"][0] - 1.0
            self.omega_matter = parameters["OmegaM0"]
            self.omega_lambda = 1.0 - self.omega_matter
            # default to 0.7 for old data format
            self.hubble_constant = parameters.get("Hubble0", 0.7)

            # use the cosmological age computed by the given cosmological parameters as the current time when COMOVING is on; cosmological age is computed by subtracting the lookback time at self.current_redshift from that at z = 1e6 (i.e., very early universe)
            cosmo = Cosmology(
                hubble_constant=self.hubble_constant,
                omega_matter=self.omega_matter,
                omega_lambda=self.omega_lambda,
            )
            self.current_time = cosmo.lookback_time(self.current_redshift, 1e6)
        else:
            self.cosmological_simulation = 0
            self.current_redshift = 0.0
            self.omega_matter = 0.0
            self.omega_lambda = 0.0
            self.hubble_constant = 0.0

            # use parameters["Time"][0] as current time when COMOVING is off
            self.current_time = parameters["Time"][0]

        # make aliases to some frequently used variables
        if parameters["Model"] == "Hydro":
            self.gamma = parameters["Gamma"]
            self.eos = parameters.get("EoS", 1)  # Assume gamma-law by default
            # default to 0.6 for old data format
            self.mu = parameters.get(
                "MolecularWeight", 0.6
            )  # Assume ionized primordial by default
            self.mhd = parameters.get("Magnetohydrodynamics", 0)
            self.srhd = parameters.get("SRHydrodynamics", 0)
        else:
            self.mhd = 0
            self.srhd = 0

        # old data format (version < 2210) did not contain any information of code units
        self.parameters.setdefault("Opt__Unit", 0)

        self.geometry = geometry_parameters[parameters.get("Coordinate", 1)]

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        try:
            # define a unique way to identify GAMER datasets
            f = HDF5FileHandler(filename)
            if "Info" in f["/"].keys() and "KeyInfo" in f["/Info"].keys():
                return True
        except Exception:
            pass
        return False



import pytest

from yt.utilities.on_demand_imports import OnDemand, safe_import


def test_access_available_module():
    class os_imports(OnDemand):
        @safe_import
        def path(self):
            from os import path

            return path

    _os = os_imports()

    _os.path.join("eggs", "saussage")


def test_access_unavailable_module():
    class Bacon_imports(OnDemand):
        @safe_import
        def spam(self):
            from Bacon import spam

            return spam

    _bacon = Bacon_imports()
    with pytest.raises(
        ImportError,
        match=r"This functionality requires the Bacon package to be installed\.",
    ):
        _bacon.spam()


def test_class_invalidation():
    with pytest.raises(
        TypeError, match="class .*'s name needs to be suffixed '_imports'"
    ):

        class Bacon(OnDemand):
            pass


def test_base_class_instanciation():
    with pytest.raises(
        TypeError, match="The OnDemand base class cannot be instanciated."
    ):
        OnDemand()



from nose.tools import assert_raises
from numpy.testing import assert_equal

from yt import YTQuantity
from yt.testing import fake_random_ds


def test_bad_disk_input():
    # Fixes 1768
    ds = fake_random_ds(16)

    # Test invalid 3d array
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1, 1], (10, "kpc"), (20, "kpc"))
    desired = "Expected an array of size (3,), received 'list' of length 4"
    assert_equal(str(ex.exception), desired)

    # Test invalid float
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1], ds.domain_center, (20, "kpc"))
    desired = (
        "Expected a numeric value (or size-1 array),"
        " received 'unyt.array.unyt_array' of length 3"
    )
    assert_equal(str(ex.exception), desired)

    # Test invalid float
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1], (10, 10), (20, "kpc"))
    desired = (
        "Expected a numeric value (or tuple of format (float, String)),"
        " received an inconsistent tuple '(10, 10)'."
    )
    assert_equal(str(ex.exception), desired)

    # Test invalid iterable
    with assert_raises(TypeError) as ex:
        ds.disk(
            ds.domain_center,
            [0, 0, 1],
            (10, "kpc"),
            (20, "kpc"),
            fields=YTQuantity(1, "kpc"),
        )
    desired = "Expected an iterable object, received 'unyt.array.unyt_quantity'"
    assert_equal(str(ex.exception), desired)

    # Test invalid object
    with assert_raises(TypeError) as ex:
        ds.disk(ds.domain_center, [0, 0, 1], (10, "kpc"), (20, "kpc"), ds=ds.all_data())
    desired = (
        "Expected an object of 'yt.data_objects.static_output.Dataset' "
        "type, received "
        "'yt.data_objects.selection_objects.region.YTRegion'"
    )
    assert_equal(str(ex.exception), desired)

    # Test valid disk
    ds.disk(ds.domain_center, [0, 0, 1], (10, "kpc"), (20, "kpc"))
    ds.disk(ds.domain_center, [0, 0, 1], 10, (20, "kpc"))



import pytest

from yt.config import ytcfg
from yt.testing import fake_random_ds, requires_module_pytest as requires_module


@pytest.fixture
def within_testing():
    old_value = ytcfg["yt", "internals", "within_testing"]
    ytcfg["yt", "internals", "within_testing"] = True
    yield
    ytcfg["yt", "internals", "within_testing"] = old_value


@pytest.mark.usefixtures("within_testing")
@requires_module("glue", "astropy")
def test_glue_data_object():
    ds = fake_random_ds(16)
    ad = ds.all_data()
    ad.to_glue([("gas", "density")])



import os
import sys
import time
import uuid
import weakref
from collections import UserDict
from itertools import chain, product, repeat
from numbers import Number as numeric_type
from typing import Type

import numpy as np
from more_itertools import always_iterable

from yt.data_objects.field_data import YTFieldData
from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.index_subobjects.octree_subset import OctreeSubset
from yt.data_objects.index_subobjects.stretched_grid import StretchedGrid
from yt.data_objects.index_subobjects.unstructured_mesh import (
    SemiStructuredMesh,
    UnstructuredMesh,
)
from yt.data_objects.particle_unions import ParticleUnion
from yt.data_objects.static_output import Dataset, ParticleFile
from yt.data_objects.unions import MeshUnion
from yt.frontends.sph.data_structures import SPHParticleIndex
from yt.geometry.geometry_handler import Index, YTDataChunk
from yt.geometry.grid_geometry_handler import GridIndex
from yt.geometry.oct_container import OctreeContainer
from yt.geometry.oct_geometry_handler import OctreeIndex
from yt.geometry.unstructured_mesh_handler import UnstructuredIndex
from yt.units import YTQuantity
from yt.utilities.io_handler import io_registry
from yt.utilities.lib.cykdtree import PyKDTree
from yt.utilities.lib.misc_utilities import (
    _obtain_coords_and_widths,
    get_box_grids_level,
)
from yt.utilities.lib.particle_kdtree_tools import (
    estimate_density,
    generate_smoothing_length,
)
from yt.utilities.logger import ytLogger as mylog

from .definitions import process_data, set_particle_types
from .fields import StreamFieldInfo

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from yt._maintenance.backports import cached_property


class StreamGrid(AMRGridPatch):
    """
    Class representing a single In-memory Grid instance.
    """

    __slots__ = ["proc_num"]
    _id_offset = 0

    def __init__(self, id, index):
        """
        Returns an instance of StreamGrid with *id*, associated with *filename*
        and *index*.
        """
        # All of the field parameters will be passed to us as needed.
        AMRGridPatch.__init__(self, id, filename=None, index=index)
        self._children_ids = []
        self._parent_id = -1
        self.Level = -1

    def set_filename(self, filename):
        pass

    @property
    def Parent(self):
        if self._parent_id == -1:
            return None
        return self.index.grids[self._parent_id - self._id_offset]

    @property
    def Children(self):
        return [self.index.grids[cid - self._id_offset] for cid in self._children_ids]


class StreamStretchedGrid(StretchedGrid):
    _id_offset = 0

    def __init__(self, id, index):
        cell_widths = index.grid_cell_widths[id - self._id_offset]
        super().__init__(id, cell_widths, index=index)
        self._children_ids = []
        self._parent_id = -1
        self.Level = -1

    @property
    def Parent(self):
        if self._parent_id == -1:
            return None
        return self.index.grids[self._parent_id - self._id_offset]

    @property
    def Children(self):
        return [self.index.grids[cid - self._id_offset] for cid in self._children_ids]


class StreamHandler:
    def __init__(
        self,
        left_edges,
        right_edges,
        dimensions,
        levels,
        parent_ids,
        particle_count,
        processor_ids,
        fields,
        field_units,
        code_units,
        io=None,
        particle_types=None,
        periodicity=(True, True, True),
        *,
        cell_widths=None,
        parameters=None,
    ):
        if particle_types is None:
            particle_types = {}
        self.left_edges = np.array(left_edges)
        self.right_edges = np.array(right_edges)
        self.dimensions = dimensions
        self.levels = levels
        self.parent_ids = parent_ids
        self.particle_count = particle_count
        self.processor_ids = processor_ids
        self.num_grids = self.levels.size
        self.fields = fields
        self.field_units = field_units
        self.code_units = code_units
        self.io = io
        self.particle_types = particle_types
        self.periodicity = periodicity
        self.cell_widths = cell_widths

        if parameters is None:
            self.parameters = {}
        else:
            self.parameters = parameters.copy()

    def get_fields(self):
        return self.fields.all_fields

    def get_particle_type(self, field):

        if field in self.particle_types:
            return self.particle_types[field]
        else:
            return False


class StreamHierarchy(GridIndex):

    grid = StreamGrid

    def __init__(self, ds, dataset_type=None):
        self.dataset_type = dataset_type
        self.float_type = "float64"
        self.dataset = weakref.proxy(ds)  # for _obtain_enzo
        self.stream_handler = ds.stream_handler
        self.float_type = "float64"
        self.directory = os.getcwd()
        GridIndex.__init__(self, ds, dataset_type)

    def _count_grids(self):
        self.num_grids = self.stream_handler.num_grids

    def _icoords_to_fcoords(self, icoords, ires, axes=None):
        """
        We check here that we have cell_widths, and if we do, we will provide them.
        """
        if self.grid_cell_widths is None:
            return super()._icoords_to_fcoords(icoords, ires, axes)
        if axes is None:
            axes = [0, 1, 2]
        # Transpose these by reversing the shape
        coords = np.empty(icoords.shape, dtype="f8")
        cell_widths = np.empty(icoords.shape, dtype="f8")
        for i, ax in enumerate(axes):
            coords[:, i], cell_widths[:, i] = _obtain_coords_and_widths(
                icoords[:, i],
                ires,
                self.grid_cell_widths[0][ax],
                self.ds.domain_left_edge[ax].d,
            )
        return coords, cell_widths

    def _parse_index(self):
        self.grid_dimensions = self.stream_handler.dimensions
        self.grid_left_edge[:] = self.stream_handler.left_edges
        self.grid_right_edge[:] = self.stream_handler.right_edges
        self.grid_levels[:] = self.stream_handler.levels
        self.min_level = self.grid_levels.min()
        self.grid_procs = self.stream_handler.processor_ids
        self.grid_particle_count[:] = self.stream_handler.particle_count
        if self.stream_handler.cell_widths is not None:
            self.grid_cell_widths = self.stream_handler.cell_widths[:]
            self.grid = StreamStretchedGrid
        else:
            self.grid_cell_widths = None
        mylog.debug("Copying reverse tree")
        self.grids = []
        # We enumerate, so it's 0-indexed id and 1-indexed pid
        for id in range(self.num_grids):
            self.grids.append(self.grid(id, self))
            self.grids[id].Level = self.grid_levels[id, 0]
        parent_ids = self.stream_handler.parent_ids
        if parent_ids is not None:
            reverse_tree = self.stream_handler.parent_ids.tolist()
            # Initial setup:
            for gid, pid in enumerate(reverse_tree):
                if pid >= 0:
                    self.grids[gid]._parent_id = pid
                    self.grids[pid]._children_ids.append(self.grids[gid].id)
        else:
            mylog.debug("Reconstructing parent-child relationships")
            self._reconstruct_parent_child()
        self.max_level = self.grid_levels.max()
        mylog.debug("Preparing grids")
        temp_grids = np.empty(self.num_grids, dtype="object")
        for i, grid in enumerate(self.grids):
            if (i % 1e4) == 0:
                mylog.debug("Prepared % 7i / % 7i grids", i, self.num_grids)
            grid.filename = None
            grid._prepare_grid()
            grid._setup_dx()
            grid.proc_num = self.grid_procs[i]
            temp_grids[i] = grid
        self.grids = temp_grids
        mylog.debug("Prepared")

    def _reconstruct_parent_child(self):
        mask = np.empty(len(self.grids), dtype="int32")
        mylog.debug("First pass; identifying child grids")
        for i, grid in enumerate(self.grids):
            get_box_grids_level(
                self.grid_left_edge[i, :],
                self.grid_right_edge[i, :],
                self.grid_levels[i] + 1,
                self.grid_left_edge,
                self.grid_right_edge,
                self.grid_levels,
                mask,
            )
            ids = np.where(mask.astype("bool"))
            grid._children_ids = ids[0]  # where is a tuple
        mylog.debug("Second pass; identifying parents")
        self.stream_handler.parent_ids = (
            np.zeros(self.stream_handler.num_grids, "int64") - 1
        )
        for i, grid in enumerate(self.grids):  # Second pass
            for child in grid.Children:
                child._parent_id = i
                # _id_offset = 0
                self.stream_handler.parent_ids[child.id] = i

    def _initialize_grid_arrays(self):
        GridIndex._initialize_grid_arrays(self)
        self.grid_procs = np.zeros((self.num_grids, 1), "int32")

    def _detect_output_fields(self):
        # NOTE: Because particle unions add to the actual field list, without
        # having the keys in the field list itself, we need to double check
        # here.
        fl = set(self.stream_handler.get_fields())
        fl.update(set(getattr(self, "field_list", [])))
        self.field_list = list(fl)

    def _populate_grid_objects(self):
        for g in self.grids:
            g._setup_dx()
        self.max_level = self.grid_levels.max()

    def _setup_data_io(self):
        if self.stream_handler.io is not None:
            self.io = self.stream_handler.io
        else:
            self.io = io_registry[self.dataset_type](self.ds)

    def _reset_particle_count(self):
        self.grid_particle_count[:] = self.stream_handler.particle_count
        for i, grid in enumerate(self.grids):
            grid.NumberOfParticles = self.grid_particle_count[i, 0]

    def update_data(self, data):
        """
        Update the stream data with a new data dict. If fields already exist,
        they will be replaced, but if they do not, they will be added. Fields
        already in the stream but not part of the data dict will be left
        alone.
        """
        particle_types = set_particle_types(data[0])

        self.stream_handler.particle_types.update(particle_types)
        self.ds._find_particle_types()

        for i, grid in enumerate(self.grids):
            field_units, gdata, number_of_particles = process_data(data[i])
            self.stream_handler.particle_count[i] = number_of_particles
            self.stream_handler.field_units.update(field_units)
            for field in gdata:
                if field in grid.field_data:
                    grid.field_data.pop(field, None)
                self.stream_handler.fields[grid.id][field] = gdata[field]

        self._reset_particle_count()
        # We only want to create a superset of fields here.
        for field in self.ds.field_list:
            if field[0] == "all":
                self.ds.field_list.remove(field)
        self._detect_output_fields()
        self.ds.create_field_info()
        mylog.debug("Creating Particle Union 'all'")
        pu = ParticleUnion("all", list(self.ds.particle_types_raw))
        self.ds.add_particle_union(pu)
        self.ds.particle_types = tuple(set(self.ds.particle_types))


class StreamDataset(Dataset):
    _index_class: Type[Index] = StreamHierarchy
    _field_info_class = StreamFieldInfo
    _dataset_type = "stream"

    def __init__(
        self,
        stream_handler,
        storage_filename=None,
        geometry="cartesian",
        unit_system="cgs",
        default_species_fields=None,
    ):
        self.fluid_types += ("stream",)
        self.geometry = geometry
        self.stream_handler = stream_handler
        self._find_particle_types()
        name = f"InMemoryParameterFile_{uuid.uuid4().hex}"
        from yt.data_objects.static_output import _cached_datasets

        _cached_datasets[name] = self
        Dataset.__init__(
            self,
            name,
            self._dataset_type,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )

    @property
    def filename(self):
        return self.stream_handler.name

    @cached_property
    def unique_identifier(self) -> str:
        return str(self.parameters["CurrentTimeIdentifier"])

    def _parse_parameter_file(self):
        self.parameters["CurrentTimeIdentifier"] = time.time()
        self.domain_left_edge = self.stream_handler.domain_left_edge.copy()
        self.domain_right_edge = self.stream_handler.domain_right_edge.copy()
        self.refine_by = self.stream_handler.refine_by
        self.dimensionality = self.stream_handler.dimensionality
        self._periodicity = self.stream_handler.periodicity
        self.domain_dimensions = self.stream_handler.domain_dimensions
        self.current_time = self.stream_handler.simulation_time
        self.gamma = 5.0 / 3.0
        self.parameters["EOSType"] = -1
        self.parameters["CosmologyHubbleConstantNow"] = 1.0
        self.parameters["CosmologyCurrentRedshift"] = 1.0
        self.parameters["HydroMethod"] = -1
        self.parameters.update(self.stream_handler.parameters)
        if self.stream_handler.cosmology_simulation:
            self.cosmological_simulation = 1
            self.current_redshift = self.stream_handler.current_redshift
            self.omega_lambda = self.stream_handler.omega_lambda
            self.omega_matter = self.stream_handler.omega_matter
            self.hubble_constant = self.stream_handler.hubble_constant
        else:
            self.current_redshift = 0.0
            self.omega_lambda = 0.0
            self.omega_matter = 0.0
            self.hubble_constant = 0.0
            self.cosmological_simulation = 0

    def _set_units(self):
        self.field_units = self.stream_handler.field_units

    def _set_code_unit_attributes(self):
        base_units = self.stream_handler.code_units
        attrs = (
            "length_unit",
            "mass_unit",
            "time_unit",
            "velocity_unit",
            "magnetic_unit",
        )
        cgs_units = ("cm", "g", "s", "cm/s", "gauss")
        for unit, attr, cgs_unit in zip(base_units, attrs, cgs_units):
            if isinstance(unit, str):
                if unit == "code_magnetic":
                    # If no magnetic unit was explicitly specified
                    # we skip it now and take care of it at the bottom
                    continue
                else:
                    uq = self.quan(1.0, unit)
            elif isinstance(unit, numeric_type):
                uq = self.quan(unit, cgs_unit)
            elif isinstance(unit, YTQuantity):
                uq = unit
            elif isinstance(unit, tuple):
                uq = self.quan(unit[0], unit[1])
            else:
                raise RuntimeError(f"{attr} ({unit}) is invalid.")
            setattr(self, attr, uq)
        if not hasattr(self, "magnetic_unit"):
            self.magnetic_unit = np.sqrt(
                4 * np.pi * self.mass_unit / (self.time_unit**2 * self.length_unit)
            )

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        return False

    @property
    def _skip_cache(self):
        return True

    def _find_particle_types(self):
        particle_types = set()
        for k, v in self.stream_handler.particle_types.items():
            if v:
                particle_types.add(k[0])
        self.particle_types = tuple(particle_types)
        self.particle_types_raw = self.particle_types


class StreamDictFieldHandler(UserDict):
    _additional_fields = ()

    @property
    def all_fields(self):
        self_fields = chain.from_iterable(s.keys() for s in self.values())
        self_fields = list(set(self_fields))
        fields = list(self._additional_fields) + self_fields
        fields = list(set(fields))
        return fields


class StreamParticleIndex(SPHParticleIndex):
    def __init__(self, ds, dataset_type=None):
        self.stream_handler = ds.stream_handler
        super().__init__(ds, dataset_type)

    def _setup_data_io(self):
        if self.stream_handler.io is not None:
            self.io = self.stream_handler.io
        else:
            self.io = io_registry[self.dataset_type](self.ds)

    def update_data(self, data):
        """
        Update the stream data with a new data dict. If fields already exist,
        they will be replaced, but if they do not, they will be added. Fields
        already in the stream but not part of the data dict will be left
        alone.
        """
        # Alias
        ds = self.ds
        handler = ds.stream_handler

        # Preprocess
        field_units, data, _ = process_data(data)
        pdata = {}
        for key in data.keys():
            if not isinstance(key, tuple):
                field = ("io", key)
                mylog.debug("Reassigning '%s' to '%s'", key, field)
            else:
                field = key
            pdata[field] = data[key]
        data = pdata  # Drop reference count
        particle_types = set_particle_types(data)

        # Update particle types
        handler.particle_types.update(particle_types)
        ds._find_particle_types()

        # Update fields
        handler.field_units.update(field_units)
        fields = handler.fields
        for field in data.keys():
            if field not in fields._additional_fields:
                fields._additional_fields += (field,)
        fields["stream_file"].update(data)

        # Update field list
        for field in self.ds.field_list:
            if field[0] in ["all", "nbody"]:
                self.ds.field_list.remove(field)
        self._detect_output_fields()
        self.ds.create_field_info()


class StreamParticleFile(ParticleFile):
    pass


class StreamParticlesDataset(StreamDataset):
    _index_class = StreamParticleIndex
    _file_class = StreamParticleFile
    _field_info_class = StreamFieldInfo
    _dataset_type = "stream_particles"
    file_count = 1
    filename_template = "stream_file"
    _proj_type = "particle_proj"

    def __init__(
        self,
        stream_handler,
        storage_filename=None,
        geometry="cartesian",
        unit_system="cgs",
        default_species_fields=None,
    ):
        super().__init__(
            stream_handler,
            storage_filename=storage_filename,
            geometry=geometry,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        fields = list(stream_handler.fields["stream_file"].keys())
        # This is the current method of detecting SPH data.
        # This should be made more flexible in the future.
        if ("io", "density") in fields and ("io", "smoothing_length") in fields:
            self._sph_ptypes = ("io",)

    def add_sph_fields(self, n_neighbors=32, kernel="cubic", sph_ptype="io"):
        """Add SPH fields for the specified particle type.

        For a particle type with "particle_position" and "particle_mass" already
        defined, this method adds the "smoothing_length" and "density" fields.
        "smoothing_length" is computed as the distance to the nth nearest
        neighbor. "density" is computed as the SPH (gather) smoothed mass. The
        SPH fields are added only if they don't already exist.

        Parameters
        ----------
        n_neighbors : int
            The number of neighbors to use in smoothing length computation.
        kernel : str
            The kernel function to use in density estimation.
        sph_ptype : str
            The SPH particle type. Each dataset has one sph_ptype only. This
            method will overwrite existing sph_ptype of the dataset.

        """
        mylog.info("Generating SPH fields")

        # Unify units
        l_unit = "code_length"
        m_unit = "code_mass"
        d_unit = "code_mass / code_length**3"

        # Read basic fields
        ad = self.all_data()
        pos = ad[sph_ptype, "particle_position"].to(l_unit).d
        mass = ad[sph_ptype, "particle_mass"].to(m_unit).d

        # Construct k-d tree
        kdtree = PyKDTree(
            pos.astype("float64"),
            left_edge=self.domain_left_edge.to_value(l_unit),
            right_edge=self.domain_right_edge.to_value(l_unit),
            periodic=self.periodicity,
            leafsize=2 * int(n_neighbors),
        )
        order = np.argsort(kdtree.idx)

        def exists(fname):
            if (sph_ptype, fname) in self.derived_field_list:
                mylog.info(
                    "Field ('%s','%s') already exists. Skipping", sph_ptype, fname
                )
                return True
            else:
                mylog.info("Generating field ('%s','%s')", sph_ptype, fname)
                return False

        data = {}

        # Add smoothing length field
        fname = "smoothing_length"
        if not exists(fname):
            hsml = generate_smoothing_length(pos[kdtree.idx], kdtree, n_neighbors)
            hsml = hsml[order]
            data[(sph_ptype, "smoothing_length")] = (hsml, l_unit)
        else:
            hsml = ad[sph_ptype, fname].to(l_unit).d

        # Add density field
        fname = "density"
        if not exists(fname):
            dens = estimate_density(
                pos[kdtree.idx],
                mass[kdtree.idx],
                hsml[kdtree.idx],
                kdtree,
                kernel_name=kernel,
            )
            dens = dens[order]
            data[(sph_ptype, "density")] = (dens, d_unit)

        # Add fields
        self._sph_ptypes = (sph_ptype,)
        self.index.update_data(data)
        self.num_neighbors = n_neighbors


_cis = np.fromiter(
    chain.from_iterable(product([0, 1], [0, 1], [0, 1])), dtype=np.int64, count=8 * 3
)
_cis.shape = (8, 3)


def hexahedral_connectivity(xgrid, ygrid, zgrid):
    r"""Define the cell coordinates and cell neighbors of a hexahedral mesh
    for a semistructured grid. Used to specify the connectivity and
    coordinates parameters used in
    :func:`~yt.frontends.stream.data_structures.load_hexahedral_mesh`.

    Parameters
    ----------
    xgrid : array_like
       x-coordinates of boundaries of the hexahedral cells. Should be a
       one-dimensional array.
    ygrid : array_like
       y-coordinates of boundaries of the hexahedral cells. Should be a
       one-dimensional array.
    zgrid : array_like
       z-coordinates of boundaries of the hexahedral cells. Should be a
       one-dimensional array.

    Returns
    -------
    coords : array_like
        The list of (x,y,z) coordinates of the vertices of the mesh.
        Is of size (M,3) where M is the number of vertices.
    connectivity : array_like
        For each hexahedron h in the mesh, gives the index of each of h's
        neighbors. Is of size (N,8), where N is the number of hexahedra.

    Examples
    --------

    >>> xgrid = np.array([-1, -0.25, 0, 0.25, 1])
    >>> coords, conn = hexahedral_connectivity(xgrid, xgrid, xgrid)
    >>> coords
    array([[-1.  , -1.  , -1.  ],
           [-1.  , -1.  , -0.25],
           [-1.  , -1.  ,  0.  ],
           ...,
           [ 1.  ,  1.  ,  0.  ],
           [ 1.  ,  1.  ,  0.25],
           [ 1.  ,  1.  ,  1.  ]])

    >>> conn
    array([[  0,   1,   5,   6,  25,  26,  30,  31],
           [  1,   2,   6,   7,  26,  27,  31,  32],
           [  2,   3,   7,   8,  27,  28,  32,  33],
           ...,
           [ 91,  92,  96,  97, 116, 117, 121, 122],
           [ 92,  93,  97,  98, 117, 118, 122, 123],
           [ 93,  94,  98,  99, 118, 119, 123, 124]])
    """
    nx = len(xgrid)
    ny = len(ygrid)
    nz = len(zgrid)
    coords = np.zeros((nx, ny, nz, 3), dtype="float64", order="C")
    coords[:, :, :, 0] = xgrid[:, None, None]
    coords[:, :, :, 1] = ygrid[None, :, None]
    coords[:, :, :, 2] = zgrid[None, None, :]
    coords.shape = (nx * ny * nz, 3)
    cycle = np.rollaxis(np.indices((nx - 1, ny - 1, nz - 1)), 0, 4)
    cycle.shape = ((nx - 1) * (ny - 1) * (nz - 1), 3)
    off = _cis + cycle[:, np.newaxis]
    connectivity = np.array(
        ((off[:, :, 0] * ny) + off[:, :, 1]) * nz + off[:, :, 2], order="C"
    )
    return coords, connectivity


class StreamHexahedralMesh(SemiStructuredMesh):
    _connectivity_length = 8
    _index_offset = 0


class StreamHexahedralHierarchy(UnstructuredIndex):
    def __init__(self, ds, dataset_type=None):
        self.stream_handler = ds.stream_handler
        super().__init__(ds, dataset_type)

    def _initialize_mesh(self):
        coords = self.stream_handler.fields.pop("coordinates")
        connect = self.stream_handler.fields.pop("connectivity")
        self.meshes = [
            StreamHexahedralMesh(0, self.index_filename, connect, coords, self)
        ]

    def _setup_data_io(self):
        if self.stream_handler.io is not None:
            self.io = self.stream_handler.io
        else:
            self.io = io_registry[self.dataset_type](self.ds)

    def _detect_output_fields(self):
        self.field_list = list(set(self.stream_handler.get_fields()))


class StreamHexahedralDataset(StreamDataset):
    _index_class = StreamHexahedralHierarchy
    _field_info_class = StreamFieldInfo
    _dataset_type = "stream_hexahedral"


class StreamOctreeSubset(OctreeSubset):
    domain_id = 1
    _domain_offset = 1

    def __init__(self, base_region, ds, oct_handler, num_zones=2, num_ghost_zones=0):
        self._num_zones = num_zones
        self.field_data = YTFieldData()
        self.field_parameters = {}
        self.ds = ds
        self.oct_handler = oct_handler
        self._last_mask = None
        self._last_selector_id = None
        self._current_particle_type = "io"
        self._current_fluid_type = self.ds.default_fluid_type
        self.base_region = base_region
        self.base_selector = base_region.selector

        self._num_ghost_zones = num_ghost_zones

        if num_ghost_zones > 0:
            if not all(ds.periodicity):
                mylog.warning(
                    "Ghost zones will wrongly assume the domain to be periodic."
                )
            base_grid = StreamOctreeSubset(base_region, ds, oct_handler, num_zones)
            self._base_grid = base_grid

    def retrieve_ghost_zones(self, ngz, fields, smoothed=False):
        try:
            new_subset = self._subset_with_gz
            mylog.debug("Reusing previous subset with ghost zone.")
        except AttributeError:
            new_subset = StreamOctreeSubset(
                self.base_region,
                self.ds,
                self.oct_handler,
                self._num_zones,
                num_ghost_zones=ngz,
            )
            self._subset_with_gz = new_subset

        return new_subset

    def _fill_no_ghostzones(self, content, dest, selector, offset):
        # Here we get a copy of the file, which we skip through and read the
        # bits we want.
        oct_handler = self.oct_handler
        cell_count = selector.count_oct_cells(self.oct_handler, self.domain_id)
        levels, cell_inds, file_inds = self.oct_handler.file_index_octs(
            selector, self.domain_id, cell_count
        )
        levels[:] = 0
        dest.update((field, np.empty(cell_count, dtype="float64")) for field in content)
        # Make references ...
        count = oct_handler.fill_level(
            0, levels, cell_inds, file_inds, dest, content, offset
        )
        return count

    def _fill_with_ghostzones(self, content, dest, selector, offset):
        oct_handler = self.oct_handler
        ndim = self.ds.dimensionality
        cell_count = (
            selector.count_octs(self.oct_handler, self.domain_id) * self.nz**ndim
        )

        gz_cache = getattr(self, "_ghost_zone_cache", None)
        if gz_cache:
            levels, cell_inds, file_inds, domains = gz_cache
        else:
            gz_cache = (
                levels,
                cell_inds,
                file_inds,
                domains,
            ) = oct_handler.file_index_octs_with_ghost_zones(
                selector, self.domain_id, cell_count
            )
            self._ghost_zone_cache = gz_cache
        levels[:] = 0
        dest.update((field, np.empty(cell_count, dtype="float64")) for field in content)
        # Make references ...
        oct_handler.fill_level(0, levels, cell_inds, file_inds, dest, content, offset)

    def fill(self, content, dest, selector, offset):
        if self._num_ghost_zones == 0:
            return self._fill_no_ghostzones(content, dest, selector, offset)
        else:
            return self._fill_with_ghostzones(content, dest, selector, offset)


class StreamOctreeHandler(OctreeIndex):
    def __init__(self, ds, dataset_type=None):
        self.stream_handler = ds.stream_handler
        self.dataset_type = dataset_type
        super().__init__(ds, dataset_type)

    def _setup_data_io(self):
        if self.stream_handler.io is not None:
            self.io = self.stream_handler.io
        else:
            self.io = io_registry[self.dataset_type](self.ds)

    def _initialize_oct_handler(self):
        header = dict(
            dims=[1, 1, 1],
            left_edge=self.ds.domain_left_edge,
            right_edge=self.ds.domain_right_edge,
            octree=self.ds.octree_mask,
            num_zones=self.ds.num_zones,
            partial_coverage=self.ds.partial_coverage,
        )
        self.oct_handler = OctreeContainer.load_octree(header)

    def _identify_base_chunk(self, dobj):
        if getattr(dobj, "_chunk_info", None) is None:
            base_region = getattr(dobj, "base_region", dobj)
            subset = [
                StreamOctreeSubset(
                    base_region,
                    self.dataset,
                    self.oct_handler,
                    self.ds.num_zones,
                )
            ]
            dobj._chunk_info = subset
        dobj._current_chunk = list(self._chunk_all(dobj))[0]

    def _chunk_all(self, dobj):
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        yield YTDataChunk(dobj, "all", oobjs, None)

    def _chunk_spatial(self, dobj, ngz, sort=None, preload_fields=None):
        sobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        # This is where we will perform cutting of the Octree and
        # load-balancing.  That may require a specialized selector object to
        # cut based on some space-filling curve index.
        for og in sobjs:
            if ngz > 0:
                g = og.retrieve_ghost_zones(ngz, [], smoothed=True)
            else:
                g = og
            yield YTDataChunk(dobj, "spatial", [g])

    def _chunk_io(self, dobj, cache=True, local_only=False):
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for subset in oobjs:
            yield YTDataChunk(dobj, "io", [subset], None, cache=cache)

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        super()._setup_classes(dd)

    def _detect_output_fields(self):
        # NOTE: Because particle unions add to the actual field list, without
        # having the keys in the field list itself, we need to double check
        # here.
        fl = set(self.stream_handler.get_fields())
        fl.update(set(getattr(self, "field_list", [])))
        self.field_list = list(fl)


class StreamOctreeDataset(StreamDataset):
    _index_class = StreamOctreeHandler
    _field_info_class = StreamFieldInfo
    _dataset_type = "stream_octree"

    levelmax = None

    def __init__(
        self,
        stream_handler,
        storage_filename=None,
        geometry="cartesian",
        unit_system="cgs",
        default_species_fields=None,
    ):
        super().__init__(
            stream_handler,
            storage_filename,
            geometry,
            unit_system,
            default_species_fields=default_species_fields,
        )
        # Set up levelmax
        self.max_level = stream_handler.levels.max()
        self.min_level = stream_handler.levels.min()


class StreamUnstructuredMesh(UnstructuredMesh):
    _index_offset = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connectivity_length = self.connectivity_indices.shape[1]


class StreamUnstructuredIndex(UnstructuredIndex):
    def __init__(self, ds, dataset_type=None):
        self.stream_handler = ds.stream_handler
        super().__init__(ds, dataset_type)

    def _initialize_mesh(self):
        coords = self.stream_handler.fields.pop("coordinates")
        connect = always_iterable(self.stream_handler.fields.pop("connectivity"))

        self.meshes = [
            StreamUnstructuredMesh(i, self.index_filename, c1, c2, self)
            for i, (c1, c2) in enumerate(zip(connect, repeat(coords)))
        ]
        self.mesh_union = MeshUnion("mesh_union", self.meshes)

    def _setup_data_io(self):
        if self.stream_handler.io is not None:
            self.io = self.stream_handler.io
        else:
            self.io = io_registry[self.dataset_type](self.ds)

    def _detect_output_fields(self):
        self.field_list = list(set(self.stream_handler.get_fields()))
        fnames = list({fn for ft, fn in self.field_list})
        self.field_list += [("all", fname) for fname in fnames]


class StreamUnstructuredMeshDataset(StreamDataset):
    _index_class = StreamUnstructuredIndex
    _field_info_class = StreamFieldInfo
    _dataset_type = "stream_unstructured"

    def _find_particle_types(self):
        pass



import yt

# Load the dataset.
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Create a slice plot for the dataset.  With no additional arguments,
# the width will be the size of the domain and the center will be the
# center of the simulation box
slc = yt.SlicePlot(ds, "z", ("gas", "density"))

# Create a list of a couple of widths and units.
# (N.B. Mpc (megaparsec) != mpc (milliparsec)
widths = [(1, "Mpc"), (15, "kpc")]

# Loop through the list of widths and units.
for width, unit in widths:

    # Set the width.
    slc.set_width(width, unit)

    # Write out the image with a unique name.
    slc.save("%s_%010d_%s" % (ds, width, unit))

zoomFactors = [2, 4, 5]

# recreate the original slice
slc = yt.SlicePlot(ds, "z", ("gas", "density"))

for zoomFactor in zoomFactors:

    # zoom in
    slc.zoom(zoomFactor)

    # Write out the image with a unique name.
    slc.save("%s_%i" % (ds, zoomFactor))



from unyt.dimensions import dimensionless
from unyt.unit_registry import *

default_unit_registry = UnitRegistry(unit_system="cgs") # type: ignore

default_unit_registry.add("h", 1.0, dimensionless, tex_repr=r"h")



import yt
from yt.utilities.answer_testing.framework import (
    GenericImageTest,
    data_dir_load,
    requires_ds,
)


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def compare(ds, field, test_prefix, decimals=12):
    def slice_image(filename_prefix):
        sl = yt.SlicePlot(ds, "z", field)
        sl.set_log("all", False)
        image_file = sl.save(filename_prefix)
        return image_file

    slice_image.__name__ = f"slice_{test_prefix}"
    test = GenericImageTest(ds, slice_image, decimals)
    test.prefix = test_prefix
    return test


raw_fields = "Laser/plt00015"
_raw_field_names = [
    ("raw", "Bx"),
    ("raw", "By"),
    ("raw", "Bz"),
    ("raw", "Ex"),
    ("raw", "Ey"),
    ("raw", "Ez"),
    ("raw", "jx"),
    ("raw", "jy"),
    ("raw", "jz"),
]


@requires_ds(raw_fields)
def test_raw_field_slices():
    ds = data_dir_load(raw_fields)
    for field in _raw_field_names:
        yield compare(ds, field, f"answers_raw_{field[1]}")



from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer0D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import validate_3d_array, validate_object
from yt.units import YTArray


class YTPoint(YTSelectionContainer0D):
    """
    A 0-dimensional object defined by a single point

    Parameters
    ----------
    p: array_like
        A points defined within the domain.  If the domain is
        periodic its position will be corrected to lie inside
        the range [DLE,DRE) to ensure one and only one cell may
        match that point
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
        A dictionary of field parameters than can be accessed by derived
        fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5, 0.5, 0.5]
    >>> point = ds.point(c)
    """

    _type_name = "point"
    _con_args = ("p",)

    def __init__(self, p, ds=None, field_parameters=None, data_source=None):
        validate_3d_array(p)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        super().__init__(ds, field_parameters, data_source)
        if isinstance(p, YTArray):
            # we pass p through ds.arr to ensure code units are attached
            self.p = self.ds.arr(p)
        else:
            self.p = self.ds.arr(p, "code_length")



from yt.fields.field_info_container import FieldInfoContainer

# We need to specify which fields we might have in our dataset.  The field info
# container subclass here will define which fields it knows about.  There are
# optionally methods on it that get called which can be subclassed.


class CM1FieldInfo(FieldInfoContainer):
    known_other_fields = (
        # Each entry here is of the form
        # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),
        ("uinterp", ("m/s", ["velocity_x"], None)),
        ("vinterp", ("m/s", ["velocity_y"], None)),
        ("winterp", ("m/s", ["velocity_z"], None)),
        ("u", ("m/s", ["velocity_x"], None)),
        ("v", ("m/s", ["velocity_y"], None)),
        ("w", ("m/s", ["velocity_z"], None)),
        ("hwin_sr", ("m/s", ["storm_relative_horizontal_wind_speed"], None)),
        ("windmag_sr", ("m/s", ["storm_relative_3D_wind_speed"], None)),
        ("hwin_gr", ("m/s", ["ground_relative_horizontal_wind_speed"], None)),
        ("thpert", ("K", ["potential_temperature_perturbation"], None)),
        ("thrhopert", ("K", ["density_potential_temperature_perturbation"], None)),
        ("prespert", ("hPa", ["presure_perturbation"], None)),
        ("rhopert", ("kg/m**3", ["density_perturbation"], None)),
        ("dbz", ("dB", ["simulated_reflectivity"], None)),
        ("qvpert", ("g/kg", ["water_vapor_mixing_ratio_perturbation"], None)),
        ("qc", ("g/kg", ["cloud_liquid_water_mixing_ratio"], None)),
        ("qr", ("g/kg", ["rain_mixing_ratio"], None)),
        ("qi", ("g/kg", ["cloud_ice_mixing_ratio"], None)),
        ("qs", ("g/kg", ["snow_mixing_ratio"], None)),
        ("qg", ("g/kg", ["graupel_or_hail_mixing_ratio"], None)),
        ("qcloud", ("g/kg", ["sum_of_cloud_water_and_cloud_ice_mixing_ratios"], None)),
        ("qprecip", ("g/kg", ["sum_of_rain_graupel_snow_mixing_ratios"], None)),
        ("nci", ("1/cm**3", ["number_concerntration_of_cloud_ice"], None)),
        ("ncr", ("1/cm**3", ["number_concentration_of_rain"], None)),
        ("ncs", ("1/cm**3", ["number_concentration_of_snow"], None)),
        ("ncg", ("1/cm**3", ["number_concentration_of_graupel_or_hail"], None)),
        ("xvort", ("1/s", ["vorticity_x"], None)),
        ("yvort", ("1/s", ["vorticity_y"], None)),
        ("zvort", ("1/s", ["vorticity_z"], None)),
        ("hvort", ("1/s", ["horizontal_vorticity_magnitude"], None)),
        ("vortmag", ("1/s", ["vorticity_magnitude"], None)),
        ("streamvort", ("1/s", ["streamwise_vorticity"], None)),
        ("khh", ("m**2/s", ["khh"], None)),
        ("khv", ("m**2/s", ["khv"], None)),
        ("kmh", ("m**2/s", ["kmh"], None)),
        ("kmv", ("m**2/s", ["kmv"], None)),
    )

    known_particle_fields = (
        # Identical form to above
        # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),
    )

    def setup_fluid_fields(self):
        # Here we do anything that might need info about the dataset.
        # You can use self.alias, self.add_output_field (for on-disk fields)
        # and self.add_field (for derived fields).
        pass

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)
        # This will get called for every particle type.




# distutils: libraries = STD_LIBS
"""
Simple readers for fortran unformatted data, specifically for the Tiger code.



"""


import numpy as np

cimport cython
cimport libc.stdlib as stdlib
cimport numpy as np
from libc.stdio cimport FILE, fclose, fopen

#cdef inline int imax(int i0, int i1):
    #if i0 > i1: return i0
    #return i1

cdef extern from "endian_swap.h":
    void FIX_SHORT( unsigned short )
    void FIX_LONG( unsigned )
    void FIX_FLOAT( float )

cdef extern from "platform_dep.h":
    void *alloca(size_t)

cdef extern from "stdio.h":
    cdef int SEEK_SET
    cdef int SEEK_CUR
    cdef int SEEK_END
    int fseek(FILE *stream, long offset, int whence)
    size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
    long ftell(FILE *stream)
    char *fgets(char *s, int size, FILE *stream)

@cython.boundscheck(False)
@cython.wraparound(False)
def read_and_seek(char *filename, np.int64_t offset1,
                  np.int64_t offset2, np.ndarray buffer, int bytes):
    cdef FILE *f = fopen(filename, "rb")
    cdef void *buf = <void *> buffer.data
    cdef char line[1024]
    cdef size_t n = 1023
    fseek(f, offset1, SEEK_SET)
    fgets(line, n, f)
    fseek(f, offset2, SEEK_CUR)
    fread(buf, 1, bytes, f)
    fclose(f)

def count_art_octs(char *fn, long offset,
                   int min_level, int max_level,
                   int nhydro_vars,
                   level_info):
    cdef int nchild = 8
    cdef int next_record = -1, nLevel = -1
    cdef int dummy_records[9]
    cdef int readin = -1
    cdef FILE *f = fopen(fn, "rb")
    fseek(f, offset, SEEK_SET)
    for _ in range(min_level + 1, max_level + 1):
        fread(dummy_records, sizeof(int), 2, f);
        fread(&nLevel, sizeof(int), 1, f); FIX_LONG(nLevel)
        print(level_info)
        level_info.append(nLevel)
        fread(dummy_records, sizeof(int), 2, f);
        fread(&next_record, sizeof(int), 1, f); FIX_LONG(next_record)
        print("Record size is:", next_record)
        # Offset for one record header we just read
        next_record = (nLevel * (next_record + 2*sizeof(int))) - sizeof(int)
        fseek(f, next_record, SEEK_CUR)
        # Now we skip the second section
        fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
        nhydro_vars = next_record/4-2-3 #nhvar in daniel's code
        #record length is normally 2 pad bytes, 8 + 2 hvars (the 2 is nchem)
        # and then 3 vars, but we can find nhvars only here and not in other
        # file headers
        next_record = (2*sizeof(int) + readin) * (nLevel * nchild)
        next_record -= sizeof(int)
        fseek(f, next_record, SEEK_CUR)
    print("nhvars",nhydro_vars)
    fclose(f)

def read_art_tree(char *fn, long offset,
                  int min_level, int max_level,
                  np.ndarray[np.int64_t, ndim=2] oct_indices,
                  np.ndarray[np.int64_t, ndim=1] oct_levels,
                  np.ndarray[np.int64_t, ndim=2] oct_info):
    #             np.ndarray[np.int64_t, ndim=1] oct_mask,
    #             np.ndarray[np.int64_t, ndim=1] oct_parents,

    # This accepts the filename of the ART header and an integer offset that
    # points to the start of the record *following* the reading of iOctFree and
    # nOct.  For those following along at home, we only need to read:
    #   iOctPr, iOctLv
    cdef int nchild = 8
    cdef int iOct, nLevel, ic1
    cdef np.int64_t next_record = -1
    cdef long long child_record
    cdef int iOctPs[3]
    cdef np.int64_t dummy_records[9]
    cdef int readin = -1
    cdef FILE *f = fopen(fn, "rb")
    fseek(f, offset, SEEK_SET)
    cdef int Level = -1
    cdef int * iNOLL = <int *> alloca(sizeof(int)*(max_level-min_level+1))
    cdef int * iHOLL = <int *> alloca(sizeof(int)*(max_level-min_level+1))
    cdef int iOctMax = 0
    level_offsets = [0]
    for _ in range(min_level + 1, max_level + 1):
        fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
        fread(&Level, sizeof(int), 1, f); FIX_LONG(Level)
        fread(&iNOLL[Level], sizeof(int), 1, f); FIX_LONG(iNOLL[Level])
        fread(&iHOLL[Level], sizeof(int), 1, f); FIX_LONG(iHOLL[Level])
        fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
        iOct = iHOLL[Level] - 1
        nLevel = iNOLL[Level]
        #print("Reading Hierarchy for Level", Lev, Level, nLevel, iOct)
        #print(ftell(f))
        for ic1 in range(nLevel):
            iOctMax = max(iOctMax, iOct)
            #print(readin, iOct, nLevel, sizeof(int))
            next_record = ftell(f)
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            assert readin==52
            next_record += readin + sizeof(int)
            fread(iOctPs, sizeof(int), 3, f)
            FIX_LONG(iOctPs[0]); FIX_LONG(iOctPs[1]); FIX_LONG(iOctPs[2])
            oct_indices[iOct, 0] = iOctPs[0]
            oct_indices[iOct, 1] = iOctPs[1]
            oct_indices[iOct, 2] = iOctPs[2]
            oct_info[iOct, 1] = ic1
            #grid_info[iOct, 2] = iOctPr # we don't seem to need this
            fread(dummy_records, sizeof(int), 6, f) # skip Nb
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            #oct_parents[iOct] = readin - 1
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            oct_levels[iOct] = readin
            fread(&iOct, sizeof(int), 1, f); FIX_LONG(iOct);
            iOct -= 1
            assert next_record > 0
            fseek(f, next_record, SEEK_SET)
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            assert readin==52

        level_offsets.append(ftell(f))

        #skip over the hydro variables
        #find the length of one child section
        #print('measuring child record ',)
        fread(&next_record, sizeof(int), 1, f);
        #print(next_record,)
        FIX_LONG(next_record)
        #print(next_record)
        fseek(f,ftell(f)-sizeof(int),SEEK_SET) #rewind
        #This is a sloppy fix; next_record is 64bit
        #and I don't think FIX_LONG(next_record) is working
        #correctly for 64bits
        if next_record > 4294967296L:
            next_record -= 4294967296L
        assert next_record == 56

        #find the length of all of the children section
        child_record = ftell(f) +  (next_record+2*sizeof(int))*nLevel*nchild
        #print('Skipping over hydro vars', ftell(f), child_record)
        fseek(f, child_record, SEEK_SET)

        # for ic1 in range(nLevel * nchild):
        #     fread(&next_record, sizeof(int), 1, f); FIX_LONG(next_record)
        #     fread(&idc, sizeof(int), 1, f); FIX_LONG(idc); idc -= 1 + (128**3)
        #     fread(&cm, sizeof(int), 1, f); FIX_LONG(cm)
        #     #if cm == 0: oct_mask[idc] = 1
        #     #else: total_masked += 1
        #     assert next_record > 0
        #     fseek(f, next_record - sizeof(int), SEEK_CUR)
    fclose(f)
    return level_offsets

def read_art_root_vars(char *fn, long root_grid_offset,
                    int nhydro_vars, int nx, int ny, int nz,
                    int ix, int iy, int iz, fields, var):

    cdef FILE *f = fopen(fn, "rb")
    cdef int j,l, cell_record_size = nhydro_vars * sizeof(float)
    cdef float temp = -1
    l=0
    fseek(f, root_grid_offset, SEEK_SET)
    # Now we seet out the cell we want
    cdef int my_offset = (((iz * ny) + iy) * nx + ix)
    #print(cell_record_size, my_offset, ftell(f))
    fseek(f, cell_record_size * my_offset, SEEK_CUR)
    #(((C)*GridDimension[1]+(B))*GridDimension[0]+A)
    for j in range(nhydro_vars):
        fread(&temp, sizeof(float), 1, f);
        if j in fields:
            FIX_FLOAT(temp)
            var[l]=temp
            l+=1
    fclose(f)

cdef void read_art_vars(FILE *f,
                    int min_level, int max_level, int nhydro_vars,
                    int grid_level,long grid_id,long child_offset,
                    fields,
                    np.ndarray[np.int64_t, ndim=1] level_offsets,
                    var):
    # nhydro_vars is the number of columns- 3 (adjusting for vars)
    # this is normally 10=(8+2chem species)
    cdef int record_size = 2+1+1+nhydro_vars+2
    cdef float temp = -1.0
    cdef float varpad[2]
    cdef int new_padding = -1
    cdef int padding[3]
    cdef long offset = 8*grid_id*record_size*sizeof(float)
    fseek(f, level_offsets[grid_level] + offset, SEEK_SET)
    for j in range(8): #iterate over the children
        l = 0
        fread(padding, sizeof(int), 3, f); FIX_LONG(padding[0])
        #print("Record Size", padding[0])
        # This should be replaced by an fread of nhydro_vars length
        for k in range(nhydro_vars): #iterate over the record
            fread(&temp, sizeof(float), 1, f); FIX_FLOAT(temp)
            #print(k, temp)
            if k in fields:
                var[j,l] = temp
                l += 1
        fread(varpad, sizeof(float), 2, f)
        fread(&new_padding, sizeof(int), 1, f); FIX_LONG(new_padding)
        assert(padding[0] == new_padding)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def read_art_grid(int varindex,
              np.ndarray[np.int64_t, ndim=1] start_index,
              np.ndarray[np.int32_t, ndim=1] grid_dims,
              np.ndarray[np.float32_t, ndim=3] data,
              np.ndarray[np.uint8_t, ndim=3] filled,
              np.ndarray[np.float32_t, ndim=2] level_data,
              int level, int ref_factor,
              component_grid_info):
    cdef int gi, i, j, k, grid_id
    cdef int ir, jr, kr
    cdef int offi, offj, offk, odind
    cdef np.int64_t di, dj, dk
    cdef np.ndarray[np.int64_t, ndim=1] ogrid_info
    cdef np.ndarray[np.int64_t, ndim=1] og_start_index
    cdef np.float64_t temp_data
    cdef np.int64_t end_index[3]
    cdef int kr_offset, jr_offset, ir_offset
    cdef int to_fill = 0
    # Note that indexing into a cell is:
    #   (k*2 + j)*2 + i
    for i in range(3):
        end_index[i] = start_index[i] + grid_dims[i]
    for gi in range(len(component_grid_info)):
        ogrid_info = component_grid_info[gi]
        grid_id = ogrid_info[1]
        og_start_index = ogrid_info[3:6] #the oct left edge
        for i in range(2*ref_factor):
            di = i + og_start_index[0] * ref_factor
            if di < start_index[0] or di >= end_index[0]: continue
            ir = <int> (i / ref_factor)
            for j in range(2 * ref_factor):
                dj = j + og_start_index[1] * ref_factor
                if dj < start_index[1] or dj >= end_index[1]: continue
                jr = <int> (j / ref_factor)
                for k in range(2 * ref_factor):
                    dk = k + og_start_index[2] * ref_factor
                    if dk < start_index[2] or dk >= end_index[2]: continue
                    kr = <int> (k / ref_factor)
                    offi = di - start_index[0]
                    offj = dj - start_index[1]
                    offk = dk - start_index[2]
                    #print(offi, filled.shape[0],)
                    #print(offj, filled.shape[1],)
                    #print(offk, filled.shape[2])
                    if filled[offi, offj, offk] == 1: continue
                    if level > 0:
                        odind = (kr*2 + jr)*2 + ir
                        # Replace with an ART-specific reader
                        #temp_data = local_hydro_data.m_var_array[
                        #        level][8*offset + odind]
                        temp_data = level_data[varindex, 8*grid_id + odind]
                    else:
                        kr_offset = kr + <int> (start_index[0] / ref_factor)
                        jr_offset = jr + <int> (start_index[1] / ref_factor)
                        ir_offset = ir + <int> (start_index[2] / ref_factor)
                        odind = (kr_offset * grid_dims[0] + jr_offset)*grid_dims[1] + ir_offset
                        temp_data = level_data[varindex, odind]
                    data[offi, offj, offk] = temp_data
                    filled[offi, offj, offk] = 1
                    to_fill += 1
    return to_fill

@cython.cdivision(True)
@cython.boundscheck(True)
@cython.wraparound(False)
def fill_child_mask(np.ndarray[np.int64_t, ndim=2] file_locations,
                    np.ndarray[np.int64_t, ndim=1] grid_le,
                    np.ndarray[np.uint8_t, ndim=4] art_child_masks,
                    np.ndarray[np.uint8_t, ndim=3] child_mask):

    #loop over file_locations, for each row extracting the index & LE
    #of the oct we will pull pull from art_child_masks
    #then use the art_child_masks info to fill in child_mask
    cdef int i,ioct,x,y,z
    cdef int nocts = file_locations.shape[0]
    cdef int lex,ley,lez
    for i in range(nocts):
        ioct = file_locations[i,1] #from fortran to python indexing?
        lex = file_locations[i,3] - grid_le[0] #the oct left edge x
        ley = file_locations[i,4] - grid_le[1]
        lez = file_locations[i,5] - grid_le[2]
        for x in range(2):
            for y in range(2):
                for z in range(2):
                    child_mask[lex+x,ley+y,lez+z] = art_child_masks[ioct,x,y,z]



from yt.testing import assert_allclose_units, assert_equal, fake_random_ds
from yt.utilities.chemical_formulas import ChemicalFormula
from yt.utilities.physical_ratios import _primordial_mass_fraction


def test_default_species_fields():

    # Test default case (no species fields)

    ds = fake_random_ds(32)
    sp = ds.sphere("c", (0.2, "unitary"))

    mu = 0.5924489101195808

    assert_allclose_units(mu * sp["index", "ones"], sp["gas", "mean_molecular_weight"])

    assert ("gas", "H_nuclei_density") not in ds.derived_field_list
    assert ("gas", "He_nuclei_density") not in ds.derived_field_list
    assert ("gas", "El_number_density") not in ds.derived_field_list
    assert ("gas", "H_p1_number_density") not in ds.derived_field_list
    assert ("gas", "He_p2_number_density") not in ds.derived_field_list
    assert ("gas", "H_p0_number_density") not in ds.derived_field_list
    assert ("gas", "He_p0_number_density") not in ds.derived_field_list

    # Test fully ionized case
    dsi = fake_random_ds(32, default_species_fields="ionized")
    spi = dsi.sphere("c", (0.2, "unitary"))
    amu_cgs = dsi.units.physical_constants.amu_cgs

    mueinv = 1.0 * _primordial_mass_fraction["H"] / ChemicalFormula("H").weight
    mueinv *= spi["index", "ones"]
    mueinv += 2.0 * _primordial_mass_fraction["He"] / ChemicalFormula("He").weight
    mupinv = _primordial_mass_fraction["H"] / ChemicalFormula("H").weight
    mupinv *= spi["index", "ones"]
    muainv = _primordial_mass_fraction["He"] / ChemicalFormula("He").weight
    muainv *= spi["index", "ones"]
    mueinv2 = spi["gas", "El_number_density"] * amu_cgs / spi["gas", "density"]
    mupinv2 = spi["gas", "H_p1_number_density"] * amu_cgs / spi["gas", "density"]
    muainv2 = spi["gas", "He_p2_number_density"] * amu_cgs / spi["gas", "density"]

    assert_allclose_units(mueinv, mueinv2)
    assert_allclose_units(mupinv, mupinv2)
    assert_allclose_units(muainv, muainv2)

    assert_equal(spi["gas", "H_p1_number_density"], spi["gas", "H_nuclei_density"])
    assert_equal(spi["gas", "He_p2_number_density"], spi["gas", "He_nuclei_density"])

    mu = 0.5924489101195808

    assert_allclose_units(
        mu * spi["index", "ones"], spi["gas", "mean_molecular_weight"]
    )

    # Test fully neutral case

    dsn = fake_random_ds(32, default_species_fields="neutral")
    spn = dsn.sphere("c", (0.2, "unitary"))
    amu_cgs = dsn.units.physical_constants.amu_cgs

    assert ("gas", "El_number_density") not in ds.derived_field_list

    mupinv = _primordial_mass_fraction["H"] / ChemicalFormula("H").weight
    mupinv *= spn["index", "ones"]
    muainv = _primordial_mass_fraction["He"] / ChemicalFormula("He").weight
    muainv *= spn["index", "ones"]
    mupinv2 = spn["gas", "H_p0_number_density"] * amu_cgs / spn["gas", "density"]
    muainv2 = spn["gas", "He_p0_number_density"] * amu_cgs / spn["gas", "density"]

    assert_allclose_units(mueinv, mueinv2)
    assert_allclose_units(mupinv, mupinv2)
    assert_allclose_units(muainv, muainv2)

    assert_equal(spn["gas", "H_p0_number_density"], spn["gas", "H_nuclei_density"])
    assert_equal(spn["gas", "He_p0_number_density"], spn["gas", "He_nuclei_density"])

    mu = 1.2285402715185552

    assert_allclose_units(
        mu * spn["index", "ones"], spn["gas", "mean_molecular_weight"]
    )
